"""
This module is based on ideas from the paper
"The Euclidean Algorithm Generates Traditional Musical Rhythms"
by Godfried Toussaint
"""

import matplotlib.patches as mp
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from music21 import stream
from music21.bar import Repeat
from music21.note import Note, Rest


class Euclid:
    """Euclidean Rhythm generator"""

    def __init__(self, notes: int, length: int, start=1):
        self.notes = notes
        self.length = length
        self.start = start
        self.rhythm = self._bjorklund()
        self.get_beat()

    def _bjorklund(self) -> list[int]:
        """
        Bjorklunds algorithm to generate the Euclidean rhythm
        Returns:
        """
        seq = [[1] if i < self.notes else [0] for i in range(self.length)]
        for n, i in enumerate(range(self.notes, self.length)):
            if seq[n] == [1]:
                seq[n].extend(seq[i])
        seq = seq[: -min(self.notes, (self.length - self.notes))]
        remainder = sum(len(j) < len(seq[0]) for j in seq)
        while remainder > 1:
            l = len(seq[0])
            i = 0
            for s in seq[-remainder:]:
                if i == (len(seq) - remainder):
                    break
                seq[i].extend(s)
                i += 1
            seq = seq[:-i]
            remainder = sum(len(j) < len(seq[0]) for j in seq)
        rh = []
        for el in seq:
            rh += el
        return rh

    def __repr__(self) -> str:
        return "".join(["X" if n else "." for n in self.rhythm])

    def get_beat(self, times=2) -> stream.Stream:
        self.beat = stream.Stream()
        # self.beat.insert(0, instrument.Piano())
        meas = stream.Measure()
        meas.leftBarline = Repeat(direction='start')
        meas.rightBarline = Repeat(direction='end', times=times)

        n_idx = [i for i, n in enumerate(self.rhythm) if n]
        if self.start != 1:
            s_rhythm = self.rhythm[n_idx[self.start - 1]:] + self.rhythm[:n_idx[self.start - 1]]
        else:
            s_rhythm = self.rhythm
        for i in s_rhythm:
            if i == 1:
                n = Note("A4", type="eighth")
                meas.append(n)
            else:
                r = Rest(type='eighth')
                meas.append(r)
        self.beat.append(meas)
        return self.beat

    def play(self):
        self.beat.show('midi')


class RhythmViewer:
    # Visualization attributes
    radius = 0.5

    def __init__(self, rhythm):
        self.rhythm = rhythm
        self.polygon = self._polygon()

    def _polygon(self):
        cp = mp.CirclePolygon(
            (self.radius, self.radius), radius=self.radius, resolution=len(self.rhythm), fc="y"
        )
        return cp

    def show(self, circle: bool = True, lines: bool = True):
        """
        Plots rhythms on a circle.
        Args:
            circle: Draws the circle
            lines: Connect notes with lines
        """
        fig, ax = plt.subplots(figsize=(12, 12))
        patches = [self.polygon]
        if circle:
            patches.append(plt.Circle((self.radius, self.radius), radius=self.radius, color='k'))
        verts = self.polygon.get_path().vertices[::-1]
        trans = self.polygon.get_patch_transform()
        points = trans.transform(verts)
        notes = []
        for n, pt in zip(self.rhythm, points):
            if n:
                note = plt.Circle(pt, radius=0.01, color="r")
                notes.append(pt)
                patches.append(note)
        notes += [notes[0]]

        collection = PatchCollection(patches, alpha=0.7)
        ax.add_collection(collection)
        if lines:
            ax.plot([p[0] for p in notes], [p[1] for p in notes], "k")
        ax.axis("off")


traditional_rhythms = {
    'fandango': Euclid(4, 12),
    'cueca': Euclid(2, 3),
    'khafif-e-ramal': Euclid(2, 5),
    'cumbia': Euclid(3, 4),
    'calypso': Euclid(3, 4),
    'khalif-e-saghil': Euclid(3, 4),
    'khafif-e-ramal_2': Euclid(3, 5, 2),
    'ruchenitza': Euclid(3, 7),
    'tresillo': Euclid(3, 8),
    'ruchenitza_2': Euclid(4, 7),
    'aksak': Euclid(4, 9),
    'outside_now': Euclid(4, 11),
    'york-samai': Euclid(5, 6, 2),
    'Nawakhat': Euclid(5, 7),
    'cinquillo': Euclid(5, 8),
    'spanish_tango': Euclid(5, 8, 2),
    'al-saghil-al-sani': Euclid(5, 8, 2),
    'agsag-samai': Euclid(5, 9),
    'venda': Euclid(5, 9, 2),
    'pictures_at_an_exhibition': Euclid(5, 11),
    'venda_clapping': Euclid(5, 12),
    'bossa-nova': Euclid(5, 16, 3),
    'bendir': Euclid(7, 8),
    'mpre': Euclid(7, 12),
    'samba': Euclid(7, 16, 7),
    'agogo-samba': Euclid(9, 16, 4),
    'ngbaka-maibo': Euclid(9, 16, 8),
    'aka': Euclid(11, 24, 7),
    'aka_upper_sangha': Euclid(13, 24, 4)
}



import music21 as m21

from . import circle5ths
from . import get_enharmonic
from . import scales

m21_scales = {
    "Major": m21.scale.MajorScale,
    "Minor": m21.scale.MinorScale,
    "Melodic Minor": m21.scale.MelodicMinorScale,
    "Harmonic Minor": m21.scale.HarmonicMinorScale,
    "Chromatic": m21.scale.ChromaticScale,
    "Dorian": m21.scale.DorianScale,
    "Hypoaeolian": m21.scale.HypoaeolianScale,
    "Hypodorian": m21.scale.HypodorianScale,
    "Hypophrygian": m21.scale.HypophrygianScale,
    "Hypolydian": m21.scale.HypolydianScale,
    "Hypomixolydian": m21.scale.HypomixolydianScale,
    "Hypolocrian": m21.scale.HypolocrianScale,
    "Lydian": m21.scale.LydianScale,
    "Locrian": m21.scale.LocrianScale,
    "Mixolydian": m21.scale.MixolydianScale,
    "Aeolian": m21.scale.MinorScale,
    "Octatonic": m21.scale.OctatonicScale,
    "Phrygian": m21.scale.PhrygianScale,
    "RagAsawari": m21.scale.RagAsawari,
    "RagMarwa": m21.scale.RagMarwa,
    "Weighted Hexatonic Blues": m21.scale.WeightedHexatonicBlues,
    "Whole Tone": m21.scale.WholeToneScale,
}


class Scale:
    def __init__(self, tonic: str, acc: str = "", mode: str = "Major"):
        """
        Instantiates a Scale object
        Args:
            tonic: Tonic or Root of the scale
            acc: accidental if any of the root
            mode: scale mode name.
        """
        assert (mode in scales) or (mode in m21_scales)
        assert acc in ["#", "-", ""]
        self.tonic = tonic + acc
        self.mode = mode

    def __str__(self) -> str:
        return f"{self.tonic.upper()} {self.mode}"

    def __repr__(self) -> str:
        return self.__str__()

    def get_notes(self) -> list[str]:
        """
        Returns list of scale notes
        """
        if self.mode in m21_scales:
            return [p.name for p in m21_scales[self.mode](self.tonic).pitches]
        hscale = m21.scale.ChromaticScale(self.tonic)
        return [hscale.pitches[j].name for j in scales[self.mode]]

    def as_scale(self):
        sc = m21.scale.ConcreteScale(pitches=self.get_notes())
        return sc

    @property
    def stream(self) -> m21.stream.Stream:
        """
        Scale as a music21 Stream Object for saving
        """
        return m21.stream.Stream([m21.note.Note(p) for p in self.get_notes()])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['musicir', 'musicir.harmony', 'musicir.leadsheets', 'musicir.rhythm']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy[mypy]>=1.4.35,<2.0.0',
 'attrs>=21.4.0,<22.0.0',
 'click>=8.0.1,<9.0.0',
 'matplotlib>=3.5.3,<4.0.0',
 'music21>=7.3.1,<8.0.0',
 'myst-parser>=0.17.2,<0.18.0']

entry_points = \
{'console_scripts': ['musicir = musicir.__main__:main']}

setup_kwargs = {
    'name': 'musicir',
    'version': '0.0.3',
    'description': 'Music Information Retrieval',
    'long_description': "Music Information Retrieval\n===========================\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/musicir.svg\n   :target: https://pypi.org/project/musicir/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/musicir.svg\n   :target: https://pypi.org/project/musicir/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/musicir\n   :target: https://pypi.org/project/musicir\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/musicir\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/musicir/latest.svg?label=Read%20the%20Docs\n   :target: https://musicir.readthedocs.io/\n   :alt: Read the documentation at https://musicir.readthedocs.io/\n.. |Tests| image:: https://github.com/fccoelho/musicir/workflows/Tests/badge.svg\n   :target: https://github.com/fccoelho/musicir/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/fccoelho/musicir/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/fccoelho/musicir\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\n\nFeatures\n--------\n\nMusicIR goal is to offer a simple API to accomplish common Music Information Retrieval tasks. It leverages other libraries but tries to keep the complexity low.\n\nCurrently it offers a very simple way to extract the harmony (Chord anotations) from leadsheets in `musicxml` format.\n\n\nRequirements\n------------\n\n* TODO\n\n\nInstallation\n------------\n\nYou can install *Music Information Retrieval* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install musicir\n\n\nUsage\n-----\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*Music Information Retrieval* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/fccoelho/musicir/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://musicir.readthedocs.io/en/latest/usage.html\n",
    'author': 'Flávio Codeço Coelho',
    'author_email': 'fccoelho@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fccoelho/musicir',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Command-line interface."""
from pprint import pprint

import click

from musicir.leadsheets.musicxml import SongParser


@click.command()
@click.version_option()
# @click.option('-h')
# @click.option('-m')
@click.argument("filename", type=click.File("r"))
def main(filename: str) -> None:
    """Music Information Retrieval.
    Extract the harmony from leadsheets, in musicxml format.
    returning it as a JSON array.

    FILENAME is the leadsheet from which the harmony is to be extracted.
    """
    S = SongParser(filename)
    click.echo(S.chords_as_json())


if __name__ == "__main__":
    main(prog_name="musicir")  # pragma: no cover



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import String
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class Song(Base):
    __tablename__ = "song"
    id = Column(Integer, primary_key=True)
    title = Column(String(128))
    copyright = Column(String)
    composer_id = Column(Integer, ForeignKey("musician.id"))
    composer = relationship("Musician", back_populates="songs")
    harmony = relationship("Harmony", back_populates="song")
    melody = relationship("Melody", back_populates="song")

    def __repr__(self) -> str:
        return f"Song(id={self.id!r}, title={self.title!r}, composer={self.composer!r})"


class Musician(Base):
    __tablename__ = "musician"
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    songs = relationship(
        "Song", back_populates="composer", cascade="all, delete-orphan"
    )

    def __repr__(self) -> str:
        return f"Musician(id={self.id!r}, name={self.name!r})"


class Harmony(Base):
    __tablename__ = "harmony"
    id = Column(Integer, primary_key=True)
    measures = Column(Integer)
    chords = Column(JSON)
    song_id = Column(Integer, ForeignKey("song.id"), nullable=False)
    song = relationship("Song", back_populates="harmony")

    def __repr__(self) -> str:
        return f"Harmony(id={self.id!r}), measures={self.measures!r}"


class Melody(Base):
    __tablename__ = "melody"
    id = Column(Integer, primary_key=True)
    song_id = Column(Integer, ForeignKey("song.id"), nullable=False)
    song = relationship("Song", back_populates="melody")
    notes = Column(JSON)

    def __repr__(self) -> str:
        return f"Melody(id={self.id!r}), notes={self.notes!r}"



import json
import os
import xml.etree.ElementTree as ET
from glob import glob

from music21 import converter
from numpy import full
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from .model import Base
from .model import Harmony
from .model import Melody
from .model import Musician
from .model import Song


class ChordParser:
    alt_symbols: dict[str, str] = {"-1": "♭", "1": "♯"}

    def __init__(self, element: ET.Element):
        self.root: str = element.find("root").find("root-step").text
        alterations: list[ET.Element] = [el for el in element.iter("root-alter")]
        self.alteration = "" if not len(alterations) else alterations[0].text
        self.alteration = self.alt_symbols.get(self.alteration, self.alteration)
        self.kind = element.find("kind").attrib["text"]
        self.function = element.find("kind").text

    def __repr__(self) -> str:
        return f"{self.root}{self.alteration}{self.kind} - {self.function}"

    def __str__(self) -> str:
        return self.__repr__()


class NoteParser:
    alt_symbols: dict[str, str] = {"-1": "♭", "1": "♯"}

    def __init__(self, element: ET.Element):
        pitch = element.find("pitch")
        if not pitch:
            self.is_rest = True
            self.figure = element.find("type").text
            self.duration = element.find("duration").text
            self.voice = element.find("voice").text
            self.note = None
            return
        else:
            self.is_rest = False
        self.note = pitch.find("step").text
        self.octave = pitch.find("octave").text
        try:
            self.alter = pitch.find("alter").text
            self.accidental = element.find("accidental").text
        except AttributeError:
            self.alter = ""
            self.accidental = ""
        self.figure = element.find("type").text
        self.duration = element.find("duration").text
        self.voice = element.find("voice").text

    def __repr__(self) -> str:
        if self.note:
            return f"{self.note}{self.alt_symbols.get(self.alter, '')}"
        return f"Rest {self.figure}"

    def __str__(self) -> str:
        return self.__repr__()


# @define
class SongParser:
    def __init__(self, xmlfile: str):
        """
        Extracts the harmony from musicxml file when available.
        Args:
            xmlfile: path of a musicxml file
        """
        tree = ET.parse(xmlfile)
        xmlroot = tree.getroot()
        self.measures = [measure for measure in xmlroot.iter("measure")]
        self.number_of_measures = len(self.measures)

    def get_measure_chords(self, measure: int = 0) -> list[object]:
        """
        return a list of Chords as xml elements from a specific measure.
        Args:
            measure: measure number starting at 0

        Returns: List of xml elements

        """
        meas = self.measures[measure]
        chords: list[object] = meas.findall("harmony")
        return chords

    def chords_as_json(self) -> str:
        """
        Return full harmony as a JSON array.
        Returns: JSON string

        """
        hj: list[dict[str, object]] = []
        for m in range(self.number_of_measures):
            h: list[object] = self.get_measure_chords(m)
            hj.append({"measure": m, "chords": [ChordParser(c).__str__() for c in h]})
        return json.dumps(hj)

    def melody_as_json(self) -> str:
        full_melody: list[dict[str, object]] = []
        for m in range(self.number_of_measures):
            notes: list[object] = self.get_measure_melody(m)
            full_melody.append(
                {"measure": m, "notes": [NoteParser(n).__str__() for n in notes]}
            )
        return json.dumps(full_melody)

    def get_measure_melody(self, measure: int = 0) -> list[object]:
        """
        Returns list of notes in the specified measure. Each note is an xml element
        """
        meas = self.measures[measure]
        notes: list[object] = meas.findall("note")
        return notes or []


def import_into_db(path: str) -> None:
    eng = create_engine("sqlite:///leadsheets.sqlite", echo=False, future=True)
    Base.metadata.create_all(eng)
    songs: list[str] = glob(os.path.join(path, "*.xml")) + glob(
        os.path.join(path, "*.musicxml")
    )
    with Session(eng) as session:
        objs = []
        for i, song in enumerate(songs):
            print(f"{i+1} of {len(songs)}: {song}loading {song}...\r", end="")
            SO = converter.parse(song)
            HP = SongParser(song)
            sng = Song(
                title=SO.metadata.title,
                composer=Musician(name=SO.metadata.composer),
                copyright=SO.metadata.all()[1][1],
                harmony=[
                    Harmony(measures=HP.number_of_measures, chords=HP.chords_as_json())
                ],
            )
            objs.append(sng)
        session.add_all(objs)
        session.commit()
